from __future__ import annotations

import math
from dataclasses import asdict, dataclass
from typing import Any

import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike, NDArray
from scipy.optimize import root_scalar

from cratermaker._cratermaker import simplemoon_functions as sm
from cratermaker.components.morphology import Morphology
from cratermaker.components.surface import LocalSurface, Surface
from cratermaker.constants import FloatLike
from cratermaker.core.crater import Crater
from cratermaker.utils.general_utils import format_large_units, parameter


@dataclass(frozen=True, slots=True)
class SimpleMoonCrater(Crater):
    rim_height: float | None = None
    rim_width: float | None = None
    floor_depth: float | None = None
    floor_diameter: float | None = None
    peak_height: float | None = None
    ejrim: float | None = None

    def __str__(self) -> str:
        base = super().__str__()
        return (
            f"{base}\n"
            f"Rim height: {format_large_units(self.rim_height, quantity='length')}\n"
            f"Rim width: {format_large_units(self.rim_width, quantity='length')}\n"
            f"Floor depth: {format_large_units(self.floor_depth, quantity='length')}\n"
            f"Floor diameter: {format_large_units(self.floor_diameter, quantity='length')}\n"
            f"Central peak height: {format_large_units(self.peak_height, quantity='length') if self.peak_height else 'None'}\n"
            f"Ejecta rim thickness: {format_large_units(self.ejrim, quantity='length')}\n"
        )

    @classmethod
    def maker(cls, crater: Crater | None = None, **kwargs) -> SimpleMoonCrater:
        if crater is None:
            crater = super(cls, cls).maker(**kwargs)
        base_fields = asdict(crater)
        # Remove any SimpleMoonCrater-specific fields to avoid conflicts
        morphology_fields = set(cls.__dataclass_fields__) - set(super(cls, cls).__dataclass_fields__)
        for key in morphology_fields:
            base_fields.pop(key, None)

        diameter_m = crater.final_diameter
        diameter_km = diameter_m * 1e-3

        if crater.morphology_type in ["simple", "transitional"]:
            rim_height = 0.043 * diameter_km**1.014 * 1e3
            rim_width = 0.257 * diameter_km**1.011 * 1e3
            floor_depth = 0.224 * diameter_km**1.010 * 1e3
            floor_diameter = 0.200 * diameter_km**1.143 * 1e3
            peak_height = None
        elif crater.morphology_type in ["complex", "peakring", "multiring"]:
            rim_height = 0.236 * diameter_km**0.399 * 1e3
            rim_width = 0.467 * diameter_km**0.836 * 1e3
            floor_depth = 1.044 * diameter_km**0.301 * 1e3
            floor_diameter = min(0.187 * diameter_km**1.249 * 1e3, 0.9 * diameter_m)
            peak_height = 0.032 * diameter_km**0.900 * 1e3
        else:
            raise ValueError(f"Unknown morphology type: {crater.morphology_type}")

        ejrim = 0.14 * (diameter_m * 0.5) ** 0.74

        return cls(
            **base_fields,
            rim_height=rim_height,
            rim_width=rim_width,
            floor_depth=floor_depth,
            floor_diameter=floor_diameter,
            peak_height=peak_height,
            ejrim=ejrim,
        )


@Morphology.register("simplemoon")
class SimpleMoon(Morphology):
    """
    An operations class for computing the morphology of a crater and applying it to a surface mesh.

    Parameters
    ----------
    surface : str or Surface, optional
        The name of a Surface object, or an instance of Surface, to be associated the morphology model.
    crater : Crater, optional
        The crater currently attached to the morphology model.
    ejecta_truncation : float, optional
        The relative distance from the rim of the crater to truncate the ejecta blanket, default is None, which will compute a
        truncation distance based on where the ejecta thickness reaches a small value.
    rng : Generator, optional
        A random number generator instance. If not provided, the default numpy RNG will be used.
    dorays : bool, optional
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket, default is True.
    rng : numpy.random.Generator | None
        A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        The rng_rng_seed for the RNG. If None, a new RNG is created.
    rng_state : dict, optional
        The state of the random number generator. If None, a new state is created.
    **kwargs : Any
        Additional keyword arguments.
    """

    def __init__(
        self,
        surface: Surface | str | None = None,
        ejecta_truncation: FloatLike | None = None,
        dorays: bool = True,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        object.__setattr__(self, "_ejecta_truncation", None)
        object.__setattr__(self, "_node", None)
        self.ejecta_truncation = ejecta_truncation
        self.dorays = dorays
        super().__init__(surface=surface, rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)

    def __str__(self) -> str:
        base = super().__str__()
        if self.ejecta_truncation is not None:
            base += f"\nEjecta Trunction: {self.ejecta_truncation:.2f} * crater.final_radius"
        else:
            base += "\nEjecta Truncation: Off"
        return f"{base}\nEjecta Rays: {self.dorays}"

    def emplace(self, crater: Crater, **kwargs: Any) -> None:
        """
        Convenience method to immediately emplace a crater onto the surface.

        Initializes and uses the queue system behind the scenes.

        Parameters
        ----------
        crater : Crater
            The crater to be emplaced.
        kwargs : Any
            Additional keyword arguments to pass to the emplace method.
        """
        if crater is None:
            crater_args = {**kwargs, **vars(self.common_args)}
            # Add scaling=self.scaling to the kwargs if it is not already present
            if "scaling" not in crater_args:
                crater_args["scaling"] = self.scaling
            crater = SimpleMoonCrater.maker(**crater_args)
        elif isinstance(crater, list) and len(crater) > 0:
            for i, c in enumerate(crater):
                crater[i] = Crater.maker(c)
            return
        if isinstance(crater, Crater):
            crater = SimpleMoonCrater.maker(crater)

        super().emplace(crater)
        return

    def crater_shape(self, crater: SimpleMoonCrater, region: LocalSurface, **kwargs: Any) -> NDArray[np.float64]:
        """
        Compute the crater shape based on the region view and surface.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the crater shape.
        region:  LocalSurface
            The region view of the surface mesh centered at the crater center.

        Returns
        -------
        NDArray[np.float64]
            The computed crater shape at the face and node elevations.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)
        reference_elevation = region.get_reference_surface(reference_radius=crater.final_radius)

        # Combine distances and references for nodes and faces
        distance = np.concatenate([region.face_distance, region.node_distance])

        original_elevation = np.concatenate([region.face_elevation, region.node_elevation])

        new_elevation = self.crater_profile(crater, distance, reference_elevation)
        elevation_change = new_elevation - original_elevation

        return elevation_change

    def crater_profile(self, crater: SimpleMoonCrater, r: ArrayLike, r_ref: ArrayLike | None = None) -> NDArray[np.float64]:
        """
        Compute the crater profile elevation at a given radial distance.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the crater profile.
        r : ArrayLike
            Radial distances from the crater center (in meters).
        r_ref : ArrayLike, optional
            Reference elevation values to be modified by the crater profile.

        Returns
        -------
        elevation : NDArray[np.float64]
            The computed crater elevation profile at each radial point.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)
        if r_ref is None:
            r_ref = np.zeros_like(r)

        if np.isscalar(r):
            r = np.array([r], dtype=np.float64)
        elif isinstance(r, (list | tuple)):
            r = np.array(r, dtype=np.float64)

        # flatten r to 1D array
        rflat = np.ravel(r)
        r_ref_flat = np.ravel(r_ref)
        elevation = sm.crater_profile(
            rflat,
            r_ref_flat,
            crater.final_diameter,
            crater.floor_depth,
            crater.floor_diameter,
            crater.rim_height,
            crater.ejrim,
        )
        # reshape elevation to match the shape of r
        elevation = np.array(elevation, dtype=np.float64)
        elevation = np.reshape(elevation, r.shape)

        return elevation

    def ejecta_shape(
        self, crater: SimpleMoonCrater, region: LocalSurface, **kwargs: Any
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Compute the ejecta shape based on the region view and surface.

        Parameters
        ----------
        region : LocalSurface
            The region view of the surface mesh centered at the crat er center.

        Returns
        -------
        NDArray[np.float64]
            The computed ejecta shape at the face and node elevations
        NDArray[np.float64]
            The computed ejecta intensity at the face and node elevations which is used to compute the degradation function. When `dorays` is True, this is the ray intensity function. When `dorays` is False, this is just a constant 1 over the ejecta region.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)

        distance = np.concatenate([region.face_distance, region.node_distance])
        if self.dorays:
            bearing = np.concatenate([region.face_bearing, region.node_bearing])
            thickness, intensity = self.ejecta_distribution(crater, distance, bearing)
        else:
            thickness = self.ejecta_profile(crater, distance)
            intensity = np.ones_like(thickness)

        return thickness, intensity

    def ejecta_profile(self, crater: SimpleMoonCrater, r: ArrayLike) -> NDArray[np.float64]:
        """
        Compute the ejecta elevation profile at a given radial distance.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ejecta profile.
        r : ArrayLike
            Radial distances from the crater center (in meters).

        Returns
        -------
        elevation : NDArray[np.float64]
            The computed ejecta profile at each radial point.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)
        if np.isscalar(r):
            r = np.array([r], dtype=np.float64)
        elif isinstance(r, (list | tuple)):
            r = np.array(r, dtype=np.float64)
        # flatten r to 1D array
        rflat = np.ravel(r)
        elevation = sm.ejecta_profile(rflat, crater.final_diameter, crater.ejrim)
        elevation = np.array(elevation, dtype=np.float64)
        # reshape elevation to match the shape of r
        elevation = np.reshape(elevation, r.shape)
        return elevation

    def ejecta_distribution(self, crater, r: ArrayLike, theta: ArrayLike) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Compute the ejecta thickness distribution modulated by ray patterns.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ejecta distribution.
        r : ArrayLike
            Radial distances from the crater center (in meters).
        theta : ArrayLike
            Angular bearings from the crater center (in radians).

        Returns
        -------
        thickness : NDArray[np.float64]
            The computed ejecta thickness for each (r, theta) pair.
        ray_intensity : NDArray[np.float64]
            The computed ray intensity for each (r, theta) pair.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)
        # flatten r and theta to 1D arrays
        thickness = self.ejecta_profile(crater, r)
        rflat = np.ravel(r)
        theta_flat = np.ravel(theta)
        intensity = sm.ray_intensity(
            radial_distance=rflat,
            initial_bearing=theta_flat,
            crater_diameter=crater.final_diameter,
            seed=self.rng.integers(0, 2**32 - 1),
        )
        thickness = np.array(thickness, dtype=np.float64)
        intensity = np.array(intensity, dtype=np.float64)
        thickness *= intensity
        # reshape thickness to match the shape of r and theta
        return np.reshape(thickness, r.shape), np.reshape(intensity, r.shape)

    def ray_intensity(self, crater: SimpleMoonCrater, r: ArrayLike, theta: ArrayLike) -> NDArray[np.float64]:
        """
        Compute the ray pattern intensity modulation at each (r, theta) pair.

        Parameters
        ----------
        crater : SimpleMoonCrater
            The crater object containing the parameters for the ray intensity.
        r : ArrayLike
            Radial distances from the crater center (in meters).
        theta : ArrayLike
            Angular bearings from the crater center (in radians).

        Returns
        -------
        intensity : NDArray[np.float64]
            The computed ray intensity values.

        Notes
        -----
        This is a wrapper for a compiled Rust function.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)
        # flatten r and theta to 1D arrays
        rflat = np.ravel(r)
        theta_flat = np.ravel(theta)
        intensity = sm.ray_intensity(
            rflat,
            theta_flat,
            crater.final_diameter,
            seed=self.rng.integers(0, 2**32 - 1),
        )
        intensity = np.array(intensity, dtype=np.float64)
        # reshape intensity to match the shape of r and theta
        intensity = np.reshape(intensity, r.shape)
        return intensity

    def rmax(
        self,
        crater: SimpleMoonCrater,
        minimum_thickness: FloatLike,
        feature: str = "ejecta",
    ) -> float:
        """
        Compute the maximum extent of the crater based on the minimum thickness of a feature, or the ejecta_truncation factor, whichever is smaller.

        Parameters
        ----------
        crater : Crater
            The crater object to be used. If None, the current crater object is used. If passed, the current crater object is replaced.
        minimum_thickness : FloatLike
            The minimum thickness of the feature blanket in meters.
        feature : str, optional, default = "ejecta"
            The feature to compute the maximum extent. Either "crater" or "ejecta". If "crater" is chosen, the rmax is based
            on where the raised rim is smaller than minimum thickness.

        Returns
        -------
        float
            The maximum extent of the crater or ejecta blanket in meters.
        """
        if not isinstance(crater, SimpleMoonCrater):
            crater = SimpleMoonCrater.maker(crater)

        def _profile_invert_ejecta(r):
            ans = self.ejecta_profile(crater, r) - minimum_thickness
            return ans[0]

        def _profile_invert_crater(r):
            ans = self.crater_profile(crater, r, np.zeros(1)) - minimum_thickness
            return ans[0]

        if feature == "ejecta":
            _profile_invert = _profile_invert_ejecta
        elif feature == "crater":
            _profile_invert = _profile_invert_crater
        else:
            raise ValueError("Unknown feature type. Choose either 'crater' or 'ejecta'")

        # Get the maximum extent
        lower_limit = crater.final_radius * 1.0001
        upper_limit = self.ejecta_truncation * crater.final_radius if self.ejecta_truncation else np.pi * self.surface.target.radius

        if _profile_invert(lower_limit) < 0:
            ans = lower_limit
        elif _profile_invert(upper_limit) > 0:
            ans = upper_limit
        else:
            sol = root_scalar(
                _profile_invert,
                bracket=[lower_limit, upper_limit],
                method="brentq",
            )
            ans = sol.root if sol.converged else crater.final_radius

        return float(ans)

    def degradation_function(
        self,
        final_diameter: FloatLike,
        fe: FloatLike = 100.0,
    ) -> float:
        """
        Computes the degradation function, which defines the topographic degradation that each crater contributes to the surface.

        This function implements a combination of the model by Minton et al. (2019) [#]_ for small craters and Riedel et al. (2020) [#]_ for large craters. It is currently not well-constrained, so may change in the future.

        Parameters
        ----------
        final_diameter : FloatLike
            The final diameter of the crater in meters.
        fe : FloatLike, optional
            The degradation function size factor, which is a scaling factor for the degradation function. Default is 100.0.

        Returns
        -------
        float
            The computed degradation function


        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021
        .. [#] Riedel, C., Minton, D.A., Michael, G., Orgel, C., Bogert, C.H. van der, Hiesinger, H., 2020. Degradation of Small Simple and Large Complex Lunar Craters: Not a Simple Scale Dependence. Journal of Geophysical Research: Planets 125, e2019JE006273. https://doi.org/10.1029/2019JE006273
        """

        def _kdmare(r, fe, psi):
            """
            The mare-scale degradation function from Minton et al. (2019). See eq. (32).
            """
            kv1 = 0.17
            neq1 = 0.0084
            eta = 3.2
            gamma = 2.0
            beta = 2.0
            kd1 = kv1 * (math.pi * fe**2 * neq1 * (gamma * beta / ((eta - 2.0) * (beta + gamma - eta)))) ** (-gamma / (eta - beta))
            psi = gamma * ((eta - 2.0) / (eta - beta))
            return kd1 * r**psi

        def _smooth_broken(x, a, x_break, alpha_1, alpha_2, delta):
            return a * (x / x_break) ** (alpha_1) * (0.5 * (1.0 + (x / x_break) ** (1.0 / delta))) ** ((alpha_2 - alpha_1) / delta)

        def _kd(r, fe):
            psi_1 = 2.0  # Mare scale power law exponent
            psi_2 = 1.2  # Highlands scale power law exponent
            rb = 0.5e3  # breakpoint radius
            delta = 1.0e0  # Smoothing function

            kd1 = _kdmare(rb, fe, psi_1) / (1 + (psi_1 - psi_2) / psi_1) ** 2
            return _smooth_broken(r, kd1, rb, psi_1, psi_2, delta)

        return float(_kd(final_diameter / 2, fe))

    @parameter
    def ejecta_truncation(self) -> float:
        """
        The radius at which the crater is truncated relative to the crater radius.

        Returns
        -------
        float or None
        """
        return self._ejecta_truncation

    @ejecta_truncation.setter
    def ejecta_truncation(self, value: FloatLike | None):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("truction_radius must be of type FloatLike")
            self._ejecta_truncation = float(value)
        else:
            self._ejecta_truncation = None

    @parameter
    def dorays(self) -> bool:
        """
        A flag to determine if the ray pattern should be used instead of the homogeneous ejecta blanket.

        Returns
        -------
        bool
        """
        return self._dorays

    @dorays.setter
    def dorays(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError("dorays must be of type bool")
        self._dorays = value



import re
import subprocess
from pathlib import Path

version = subprocess.check_output(["git", "describe", "--tags"]).decode().replace("v", "").strip()
# Normalize version numbers to remove leading zeros in major/minor/patch
parts = version.split(".", 2)
parts[0] = str(int(parts[0]))
parts[1] = str(int(parts[1]))
if len(parts) > 2 and "-" in parts[2]:
    subparts = parts[2].split("-", 1)
    subparts[0] = str(int(subparts[0]))
    parts[2] = "-".join(subparts)
version = ".".join(parts)
# Because Rust and Python have different versioning schemes, we need to scrub the raw version string to make it compatible with both.
# Example (develop): git tag: "v2025.05.3-alpha-41-g7a2bba8" => version "2025.5.3-a41+g7a2bba8"
# Example (release): git tag: "v2025.05.3-alpha" => version "2025.5.3-a0"
# First we try to split it along "alpha-", which is what the version has if it is a pre-release development version. .
version = version.split("alpha-")
if len(version) > 1:
    version[1] = version[1].replace("-", "+")  # replace "-" with "+" to make it compatible with Python/Rust versioning
    version = "a".join(version)
else:  # This is a release version, so there is no dash after alpha, but it needs to be changed to a0
    version = version[0].replace("alpha", "a0")

root_path = Path(__file__).resolve().parents[1]
cargo_file = root_path / "Cargo.toml"
version_file = root_path / "cratermaker" / "_version.py"

# Read Cargo.toml as text
with Path.open(cargo_file, encoding="utf-8") as f:
    cargo_contents = f.read()

# Replace the version line
new_cargo_contents = re.sub(
    r'version\s*=\s*".*?"',
    f'version = "{version}"',
    cargo_contents,
    count=1,
)

# Write back
with Path.open(cargo_file, "w", encoding="utf-8") as f:
    f.write(new_cargo_contents)
with Path.open(version_file, "w") as f:
    f.write(f'__version__ = version = "{version}"\n')

print(f"Updated Cargo.toml to version: {version}")



"""
Plot the ray intensity map for a crater
=======================================

.. rubric:: By David Minton

This example demonstrates how to create a ray intensity map for a crater using the "simplemoon" morphology model.

The ray intensity model simulates the spatial modulation of ejecta distribution in the form of rays, a feature observed around many planetary impact craters. These rays are modeled using a layered pattern of radially aligned Gaussian contributions, whose intensity decays with radial distance from the crater rim. The number and angular spread of rays vary stochastically to produce a realistic, azimuthally varying intensity field.

The model uses the ray intensity function compiled from Rust code for performance. This function computes the contribution of each ray at a given point based on its radial distance and angular bearing from the crater center. The output intensity is plotted in a 2D space normalized to the crater radius and visualized on a logarithmic color scale to emphasize the structure of the rays.

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm

from cratermaker import Crater, Morphology

crater = Crater.maker(final_radius=10.0e3)
morphology = Morphology.maker()
rc = crater.final_radius

grid_size = 1000
rmax = 20
morphology.ejecta_truncation = rmax
extent = rmax * rc
x = np.linspace(-extent, extent, grid_size)
y = np.linspace(-extent, extent, grid_size)
xx, yy = np.meshgrid(x, y)

# Convert Cartesian to polar coordinates
r = np.sqrt(xx**2 + yy**2)
theta = np.arctan2(yy, xx)

# Compute ray intensity
r_flat = r.ravel()
theta_flat = theta.ravel()
intensity_flat = morphology.ray_intensity(crater, r_flat, theta_flat)
intensity = intensity_flat.reshape(r.shape)
# Plot the intensity colormap
fig, ax = plt.subplots(figsize=(6, 5))
c = ax.imshow(
    intensity,
    norm=LogNorm(vmin=1e-4, vmax=intensity.max()),
    origin="lower",
    extent=[-rmax, rmax, -rmax, rmax],
    cmap="inferno",
)
ax.set_xlabel("x (r_c)")
ax.set_ylabel("y (r_c)")
ax.set_title("Ray Intensity Map")
fig.colorbar(c, ax=ax, label="Intensity")

plt.tight_layout()
plt.show()



import unittest

import numpy as np

from cratermaker import Projectile, Target
from cratermaker.utils.general_utils import _create_catalogue


class TestProjectile(unittest.TestCase):
    def test_sample_true_requires_mean_velocity(self):
        with self.assertRaises(ValueError):
            Projectile.maker("generic", sample=True)

    def test_sample_false_requires_velocity(self):
        with self.assertRaises(ValueError):
            Projectile.maker("generic", sample=False)

    def test_sample_false_defaults(self):
        projectile = Projectile.maker("generic", sample=False, velocity=15e3)
        self.assertEqual(projectile.velocity, 15e3)
        self.assertEqual(projectile.angle, 90.0)
        self.assertEqual(projectile.direction, 0.0)
        self.assertEqual(projectile.density, 1000.0)
        self.assertEqual(projectile.location, (0.0, 0.0))
        projectile = projectile.new_projectile()
        self.assertEqual(projectile.velocity, 15e3)
        self.assertEqual(projectile.angle, 90.0)
        self.assertEqual(projectile.direction, 0.0)
        self.assertEqual(projectile.density, 1000.0)
        self.assertEqual(projectile.location, (0.0, 0.0))

    def test_sample_false_custom_angle_direction(self):
        projectile = Projectile.maker(
            "generic",
            sample=False,
            velocity=15e3,
            angle=45.0,
            direction=180.0,
            location=(120.0, -50.0),
        )
        self.assertEqual(projectile.angle, 45.0)
        self.assertEqual(projectile.direction, 180.0)
        self.assertEqual(projectile.density, 1000.0)
        self.assertEqual(projectile.location, (120.0, -50.0))
        projectile = projectile.new_projectile()
        self.assertEqual(projectile.velocity, 15e3)
        self.assertEqual(projectile.angle, 45.0)
        self.assertEqual(projectile.direction, 180.0)
        self.assertEqual(projectile.density, 1000.0)
        self.assertEqual(projectile.location, (120.0, -50.0))

    def test_density_default_and_override(self):
        default = Projectile.maker("generic", sample=False, velocity=10e3)
        self.assertEqual(default.density, 1000.0)
        projectile = Projectile.maker(
            "generic", sample=False, velocity=10e3, density=3000.0
        )
        self.assertEqual(projectile.density, 3000.0)

        projectile = projectile.new_projectile()
        self.assertEqual(projectile.velocity, 10e3)
        self.assertEqual(projectile.angle, 90.0)
        self.assertEqual(projectile.direction, 0.0)
        self.assertEqual(projectile.density, 3000.0)

    def test_sampled_vals(self):
        popnames = ["generic", "asteroids", "comets"]
        for popname in popnames:
            projectile = Projectile.maker(popname, sample=True, mean_velocity=20e3)

            velocities = []
            angles = []
            directions = []
            lons = []
            lats = []
            Nsamples = 1000
            for _ in range(Nsamples):
                projectile = projectile.new_projectile()
                velocities.append(projectile.velocity)
                angles.append(projectile.angle)
                directions.append(projectile.direction)
                lons.append(projectile.location[0])
                lats.append(projectile.location[1])
            velocities = np.array(velocities)
            angles = np.array(angles)
            directions = np.array(directions)
            lons = np.array(lons)
            lats = np.array(lats)

            vmean = np.mean(velocities)
            anglemean = np.mean(angles)
            directionmean = np.mean(directions)

            self.assertAlmostEqual(vmean / projectile.mean_velocity, 1.0, delta=0.1)
            self.assertAlmostEqual(anglemean / 45, 1.0, delta=0.1)
            self.assertAlmostEqual(directionmean / 180, 1.0, delta=0.1)

    def test_defaults(self):
        inner = ["Mercury", "Venus", "Earth", "Moon", "Vesta", "Ceres"]
        outer = ["Europa", "Umbriel", "Titan", "Triton", "Pluto", "Charon"]
        unknown = ["Arrakis", "Salusa Secundus"]
        body_properties = [
            "name",
            "radius",
            "mass",
            "material",
            "transition_scale_type",
        ]
        body_values = [
            ("Arrakis", 5995.0e3, 4.98e24, "Sand", "silicate"),
            ("Salusa Secundus", 7200.0e3, 8.62e24, "Hard Rock", "silicate"),
        ]
        catalogue = _create_catalogue(body_properties, body_values)

        for target in inner:
            projectile = Projectile.maker(target=target)
            self.assertEqual(projectile.target.name, target)
            self.assertEqual(projectile.population, "asteroids")
        for target in outer:
            projectile = Projectile.maker(target=target)
            self.assertEqual(projectile.target.name, target)
            self.assertEqual(projectile.population, "comets")
        for planet in unknown:
            target = Target.maker(target=planet, catalogue=catalogue)
            projectile = Projectile.maker(target=target)
            self.assertEqual(projectile.target.name, planet)
            self.assertEqual(projectile.population, "generic")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from math import pi
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.typing import NDArray
from scipy.integrate import quad
from tqdm import tqdm

from cratermaker.components.production import Production
from cratermaker.constants import FloatLike
from cratermaker.core.crater import Crater
from cratermaker.utils.component_utils import ComponentBase, import_components
from cratermaker.utils.general_utils import parameter

if TYPE_CHECKING:
    from cratermaker.components.surface import LocalSurface, Surface


class Morphology(ComponentBase):
    def __init__(
        self,
        surface: Surface | str | None = None,
        production: Production | str | None = None,
        dosubpixel_degradation: bool = False,
        doslope_collapse: bool = True,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the Morphology class.

        Parameters
        ----------
        surface : str or Surface, optional
            The name of a Surface object, or an instance of Surface, to be associated the morphology model.
        production : str or Production, optional
            The name of a Production object, or an instance of Production, to be associated with the morphology model. This is used for subpixel degradation in the emplace method. It is otherwise ignored.
        dosubpixel_degradation : bool, optional
            If True, subpixel degradation will be performed during the emplacement of craters. Defaults to True.
        doslope_collapse : bool, optional
            If True, slope collapse will be performed during the emplacement of craters. Defaults to True.
        **kwargs : Any

        """
        from cratermaker.components.surface import Surface

        super().__init__(**kwargs)
        object.__setattr__(self, "_production", None)
        self._surface = Surface.maker(surface, **kwargs)
        self._queue_manager: CraterQueueManager | None = None
        if production is not None:
            self._production = Production.maker(production, **kwargs)
        self.dosubpixel_degradation = dosubpixel_degradation
        self.doslope_collapse = doslope_collapse
        return

    def __str__(self) -> str:
        base = super().__str__()
        return base

    @classmethod
    def maker(
        cls,
        morphology: str | type[Morphology] | Morphology | None = None,
        surface: Surface | str | None = None,
        production: Production | str | None = None,
        dosubpixel_degradation: bool = False,
        doslope_collapse: bool = True,
        **kwargs: Any,
    ) -> Morphology:
        """
        Initialize a component model with the given name or instance.

        Parameters
        ----------
        morphology : str or Morphology or None
            The name of the morphology model to use, or an instance of Morphology. If None, the default "simplemoon" is used.
        surface : str or Surface, optional
            The name of a Surface object, or an instance of Surface, to be associated the morphology model.
        dosubpixel_degradation : bool, optional
            If True, subpixel degradation will be performed during the emplacement of craters. Defaults to True.
        doslope_collapse : bool, optional
            If True, slope collapse will be performed during the emplacement of craters. Defaults to True.
        **kwargs : Any
            Additional keyword arguments that are required for the specific morphology model being created.

        Returns
        -------
        component
            An instance of the specified component model.

        Raises
        ------
        KeyError
            If the specified morphology model name is not found in the registry.
        TypeError
            If the specified morphology model is not a string or a subclass of Morphology.
        """
        # Call the base class version of make and pass the morphology argument as the component argument
        if morphology is None:
            morphology = "simplemoon"
        morphology = super().maker(
            component=morphology,
            surface=surface,
            production=production,
            dosubpixel_degradation=dosubpixel_degradation,
            doslope_collapse=doslope_collapse,
            **kwargs,
        )
        return morphology

    def emplace(self, crater: Crater | list[Crater], **kwargs: Any) -> None:
        """
        Convenience method to immediately emplace a crater onto the surface.

        Initializes and uses the queue system behind the scenes.

        Parameters
        ----------
        crater : Crater or list[Crater]
            The crater to be emplaced.
        **kwargs : Any
            Additional keyword arguments.
        """
        if self._queue_manager is None:
            self._init_queue_manager()

        if isinstance(crater, list) and len(crater) > 0:
            for c in crater:
                self._enqueue_crater(c)
        elif isinstance(crater, Crater):
            self._enqueue_crater(crater)
        self._process_queue()

    def form_crater(self, crater: Crater, **kwargs: Any) -> None:
        """
        Form the interior of the crater by altering the elevation variable of the surface mesh.

        Parameters
        ----------
        crater : Crater
            The crater object to be formed.
        **kwargs : dict
            Additional keyword arguments to be passed to internal functions.

        Returns
        -------
        None
        """
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")

        # Find the node and face center of the crater
        self._face_index = self.surface.find_nearest_face(crater.location)

        # Test if the ejecta is big enough to modify the surface
        ejecta_rmax = self.rmax(crater, minimum_thickness=self.surface.smallest_length, feature="ejecta")
        ejecta_region = self.surface.extract_region(crater.location, ejecta_rmax)
        ejecta_area = pi * ejecta_rmax**2
        if (
            ejecta_region is None or ejecta_area < self.surface.face_area[self.face_index]
        ):  # The crater is too small to change the surface
            return

        crater_rmax = self.rmax(crater, minimum_thickness=self.surface.smallest_length, feature="crater")
        crater_region = ejecta_region.extract_subregion(crater_rmax)
        crater_volume = None
        if crater_region is not None:  # The crater is big enough to affect the surface
            crater_area = pi * crater_rmax**2

            # Check to make sure that the face at the crater location is not smaller than the crater area
            if crater_area > self.surface.face_area[self.face_index]:
                # Form the crater shape
                elevation_change = self.crater_shape(crater, crater_region)
                crater_region.update_elevation(elevation_change)
                if self.doslope_collapse:
                    crater_region.slope_collapse()
                crater_volume = crater_region.compute_volume(elevation_change[: crater_region.n_face])

                # Remove any ejecta from the surface
                inner_crater_region = crater_region.extract_subregion(crater.final_radius)
                if inner_crater_region is not None:
                    inner_crater_region.add_data(
                        "ejecta_thickness",
                        long_name="ejecta thickness",
                        units="m",
                        data=0.0,
                        overwrite=True,
                    )

        # Now form the ejecta blanket
        ejecta_thickness, ejecta_intensity = self.ejecta_shape(crater, ejecta_region)

        if crater_volume:
            ejecta_volume = ejecta_region.compute_volume(ejecta_thickness[: ejecta_region.n_face])
            conservation_factor = -crater_volume / ejecta_volume
            ejecta_thickness *= conservation_factor

        ejecta_region.add_data(
            "ejecta_thickness",
            long_name="ejecta thickness",
            units="m",
            data=ejecta_thickness[: ejecta_region.n_face],
        )

        ejecta_region.update_elevation(ejecta_thickness)

        k_ej = self.ejecta_burial_degradation(ejecta_thickness[: ejecta_region.n_face], ejecta_soften_factor=1.50)
        ejecta_region.apply_diffusion(k_ej)

        k_deg = self.degradation_function(crater.final_diameter, fe=100) * ejecta_intensity[: ejecta_region.n_face]
        ejecta_region.apply_diffusion(k_deg)

        return

    def compute_subpixel_degradation(
        self,
        age_start: float,
        age_end: float,
        production: Production | str | None = None,
        **kwargs,
    ) -> None:
        """
        Performs the subpixel degradation.

        This models the combined degradation of the part of the production population that is below the resolution of the mesh on each face. It is called between batches of craters by the `emplace` method.

        Parameters
        ----------
        age_start : float
            The age of the surface at the start of the degradation.
        age_end : float
            The age of the surface at the end of the degradation.
        production : Production or str, optional
            The production object containing the production population that will contribute to the degradation. This is only necessary if the Morphology model was not initialized with a production object. If not provided, the production object associated with this morphology model will be used. If passed, this will override the current production object.
        **kwargs : Any
            Additional keyword arguments for the degradation function.
        """
        dc_min = 1e-8  # Minimum crater size for subpixel degradation calculation.

        if age_end >= age_start:
            raise ValueError("age_end must be less than age_start.")

        production = Production.maker(production, **kwargs) if production is not None else self.production

        if not hasattr(self, "_Kdiff"):
            self._Kdiff = np.zeros_like(self.surface.face_elevation)

        def _subpixel_degradation(final_diameter):
            fe = 100.0
            k = self.degradation_function(final_diameter, fe)
            n = production.function(
                diameter=final_diameter,
                age=age_start,
                age_end=age_end,
                validate_inputs=False,
            ).item()
            degradation_region_area = np.pi * (final_diameter / 2) * fe
            return k * n * degradation_region_area

        for face_indices, dc_max in zip(self.surface.face_bin_indices, self.surface.face_bin_max_sizes, strict=False):
            delta_kdiff, _ = quad(_subpixel_degradation, dc_min, dc_max)
            self._Kdiff[face_indices] += delta_kdiff

        # If any Kdiff values reaches a threshold where a meaningful amount of diffusion will occur on the surface, then go ahead and apply it.
        # Otherwise, degradation will continue to accumulate until the next batch of craters is processed.
        if np.any(self._Kdiff / self.surface.face_area > 1):
            self.apply_subpixel_degradation()

        return

    def apply_subpixel_degradation(self) -> None:
        """
        Apply subpixel degradation to the surface using the current Kdiff values.

        This method is called after all craters have been processed and is used to
        apply the accumulated degradation effects.
        """
        if not hasattr(self, "_Kdiff"):
            raise RuntimeError("Kdiff has not been initialized. Call compute_subpixel_degradation first.")

        self.surface.apply_diffusion(self._Kdiff)
        self._Kdiff = np.zeros_like(self.surface.face_elevation)
        return

    def _affected_indices(self, crater: Crater) -> tuple[set[int], set[int]]:
        """
        Determine the set of node and face indices affected by a crater's ejecta blanket.

        Parameters
        ----------
        crater : Crater
            The crater object whose effect region is to be computed.

        Returns
        -------
        affected_nodes : set of int
            The set of node indices affected by the crater.
        affected_faces : set of int
            The set of face indices affected by the crater.
        """
        rmax = self.rmax(crater, minimum_thickness=self.surface.smallest_length, feature="ejecta")
        region = self.surface.extract_region(crater.location, rmax)
        if region is None:
            return set(), set()
        if isinstance(region.node_indices, slice) or isinstance(region.face_indices, slice):
            return set(np.arange(self.surface.n_node)[region.node_indices]), set(
                np.arange(self.surface.n_face)[region.face_indices]
            )
        return set(region.node_indices), set(region.face_indices)

    def _init_queue_manager(self) -> None:
        """
        Initialize the crater queue manager with a surface-dependent overlap function.
        """

        def overlap_fn(crater: Crater) -> tuple[set[int], set[int]]:
            return self._affected_indices(crater)

        self._queue_manager = CraterQueueManager(overlap_fn)

    def _enqueue_crater(
        self,
        crater: Crater | None = None,
        **kwarg: Any,
    ) -> None:
        """
        Add a crater to the queue for later emplacement.

        Automatically initializes the queue manager if it hasn't been set.

        Parameters
        ----------
        crater : Crater, optional
            The crater object to enqueue. If None, one is created from keyword args.
        **kwarg : Any
            Additional keyword arguments for crater construction.

        Raises
        ------
        RuntimeError
            If the queue manager must be initialized but no surface is provided.
        """
        if self._queue_manager is None:
            if self.surface is None:
                raise RuntimeError("Surface must be provided to initialize queue manager.")
            self._init_queue_manager()

        if crater is None:
            crater = Crater.maker(**kwarg)
        self._queue_manager.push(crater)

    def _process_queue(self) -> None:
        """
        Process all queued craters in the order they were added, forming non-overlapping batches and applying each to the surface.

        Raises
        ------
        RuntimeError
            If the queue manager has not been initialized.
        """
        if not hasattr(self, "_queue_manager"):
            raise RuntimeError("Queue manager has not been initialized. Call _init_queue_manager first.")

        from concurrent.futures import ThreadPoolExecutor

        def _batch_process(pbar=None):
            while not self._queue_manager.is_empty():
                batch = self._queue_manager.peek_next_batch()

                def process(crater):
                    try:
                        self.form_crater(crater)
                        if pbar is not None:
                            pbar.update(1)
                    except Exception as e:
                        raise RuntimeError(f"Error processing crater {crater}: {e}") from e

                # max_workers=1 because something needs access to HDF files (probably grid.nc) that is not thread safe
                with ThreadPoolExecutor(max_workers=1) as executor:
                    executor.map(process, batch)

                if self.dosubpixel_degradation and len(batch) > 1:
                    # If the craters have age values attached to them, we can perform subpixel degradation between time values
                    agevals = [crater.age for crater in batch if crater.age is not None]
                    if len(agevals) > 1:
                        self.compute_subpixel_degradation(age_start=max(agevals), age_end=min(agevals))

                self._queue_manager.pop_batch(batch)
                self._queue_manager.clear_active()

            if self.dosubpixel_degradation:
                self.apply_subpixel_degradation()
            return

        total_craters = len(self._queue_manager._queue)
        if total_craters > 10:
            with tqdm(
                total=total_craters,
                desc="Processing craters",
                position=0,
                leave=False,
                unit="craters",
            ) as pbar:
                _batch_process(pbar)
        else:
            _batch_process()
        return

    def ejecta_burial_degradation(self, ejecta_thickness, ejecta_soften_factor=1.50) -> NDArray[np.float64]:
        """
        Computes the change in degradation state due to ejecta burial.

        This function implements a combination of the model by Minton et al. (2019) [#]_.

        Parameters
        ----------
        region : LocalSurface
            The region view of the surface mesh centered at the crater center.
        ejecta_thickness : NDArray[np.float64]
            The computed ejecta thickness at the face and node elevations.

        Returns
        -------
        NDArray[np.float64]
            The computed change in degradation state for all faces in the

        References
        ----------
        .. [#] Minton, D.A., Fassett, C.I., Hirabayashi, M., Howl, B.A., Richardson, J.E., (2019). The equilibrium size-frequency distribution of small craters reveals the effects of distal ejecta on lunar landscape morphology. Icarus 326, 63-87. https://doi.org/10.1016/j.icarus.2019.02.021
        """
        return ejecta_soften_factor * ejecta_thickness**2

    @abstractmethod
    def degradation_function(self, final_diameter: FloatLike, fe: FloatLike) -> float: ...

    @abstractmethod
    def crater_shape(
        self,
        crater: Crater,
        region: LocalSurface,
        **kwarg: Any,
    ) -> NDArray[np.float64]: ...

    @abstractmethod
    def ejecta_shape(
        self,
        crater: Crater,
        region: LocalSurface,
        **kwarg: Any,
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...

    @abstractmethod
    def rmax(
        self,
        crater: Crater,
        minimum_thickness: FloatLike,
        feature: str = "ejecta",
    ) -> float: ...

    @property
    def surface(self) -> Surface:
        """
        The surface object associated with this morphology model.
        """
        return self._surface

    @surface.setter
    def surface(self, surface: Surface) -> None:
        """
        Set the surface object associated with this morphology model.
        """
        from cratermaker.components.surface import Surface

        if not isinstance(surface, (Surface | str)):
            raise TypeError("surface must be an instance of Surface or a string")
        self._surface = Surface.maker(surface)
        self._queue_manager: CraterQueueManager | None = None

    @property
    def production(self) -> Production:
        """
        The production object associated with this morphology model.
        """
        return self._production

    @production.setter
    def production(self, production: Production) -> None:
        """
        Set the production object associated with this morphology model.
        """
        from cratermaker.components.production import Production

        if not isinstance(production, (Production | str)):
            raise TypeError("production must be an instance of Production or a string")
        self._production = Production.maker(production)

    @parameter
    def dosubpixel_degradation(self) -> bool:
        """
        Whether to perform subpixel degradation during crater emplacement.
        """
        return self._dosubpixel_degradation

    @dosubpixel_degradation.setter
    def dosubpixel_degradation(self, value: bool) -> None:
        """
        Set whether to perform subpixel degradation during crater emplacement.
        """
        if not isinstance(value, bool):
            raise TypeError("dosubpixel_degradation must be a boolean value")
        self._dosubpixel_degradation = value

    @parameter
    def doslope_collapse(self) -> bool:
        """
        Whether to perform slope collapse during crater emplacement.
        """
        return self._doslope_collapse

    @doslope_collapse.setter
    def doslope_collapse(self, value: bool) -> None:
        """
        Set whether to perform slope collapse during crater emplacement.
        """
        if not isinstance(value, bool):
            raise TypeError("doslope_collapse must be a boolean value")
        self._doslope_collapse = value

    @property
    def face_index(self):
        """
        The index of the face closest to the crater location.

        Returns
        -------
        int
        """
        return self._face_index


class CraterQueueManager:
    """
    A manager for craters awaiting emplacement. Craters are processed in order (FIFO) but batches of non-overlapping craters can be processed simultaneously.

    Parameters
    ----------
    overlap_fn : Callable[[Crater], tuple[set[int], set[int]]]
        A function that takes a crater and returns a tuple of (node indices, face indices) affected.
    """

    def __init__(self, overlap_fn: Callable[[Crater], tuple[set[int], set[int]]]):
        self._queue: list[Crater] = []
        self._active_nodes: set[int] = set()
        self._active_faces: set[int] = set()
        self._overlap_fn = overlap_fn

    def push(self, crater: Crater) -> None:
        self._queue.append(crater)

    def peek_next_batch(self) -> list[Crater]:
        """
        Return a list of the next batch of craters that do not overlap with each other or the current active region.
        """
        batch = []
        reserved_nodes = set(self._active_nodes)
        reserved_faces = set(self._active_faces)
        for crater in self._queue:
            node_indices, face_indices = self._overlap_fn(crater)
            if reserved_nodes.isdisjoint(node_indices) and reserved_faces.isdisjoint(face_indices):
                batch.append(crater)
                reserved_nodes.update(node_indices)
                reserved_faces.update(face_indices)
            else:
                break
        return batch

    def pop_batch(self, batch: list[Crater]) -> None:
        """
        Remove a processed batch of craters from the queue.
        """
        for crater in batch:
            self._queue.remove(crater)
            node_indices, face_indices = self._overlap_fn(crater)
            self._active_nodes.update(node_indices)
            self._active_faces.update(face_indices)

    def clear_active(self) -> None:
        """
        Clear the active region set after batch processing is complete.
        """
        self._active_nodes.clear()
        self._active_faces.clear()

    def is_empty(self) -> bool:
        return len(self._queue) == 0


import_components(__name__, __path__, ignore_private=True)



"""
Plot the Neukum projectile CSFD
========================================

.. rubric:: By Austin Blevins and David Minton

In this example, we will plot the cumulative size-frequency distribution of projectiles for a 1 Gy age, which is derived from Ivanov, Neukum, and Wagner (2001) [#]_.

References
----------

.. [#] Ivanov, B.A., Neukum, G., Wagner, R., 2001. Size-Frequency Distributions of Planetary Impact Craters and Asteroids. In *Collisional Processes in the Solar System*, Springer Netherlands, Dordrecht, pp. 1-34.  https://doi.org/10.1007/978-94-010-0712-2_1
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import ticker

from cratermaker import Production

production = Production.maker("neukum", version="Projectile")
Dvals = np.logspace(-2, 7, num=1000)
Nvals = production.function(diameter=Dvals, age=1000.0)

fig, ax = plt.subplots(figsize=(4, 7))
Dvals *= 1e-3  # Convert from m to km
Nvals *= 1e6  # Convert from m^2 to km^2
x_min = Dvals[0]
x_max = Dvals[-1]
ax.set_xscale("log")
ax.set_yscale("log")
ax.set_ylabel("$\\mathregular{N_{>D}}$")
ax.set_xlabel("Projectile Diameter (km)")
ax.set_xlim(x_min, x_max)
ax.yaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
ax.yaxis.set_minor_locator(
    ticker.LogLocator(base=10.0, subs=np.arange(2, 10), numticks=100)
)
ax.xaxis.set_major_locator(ticker.LogLocator(base=10.0, numticks=20))
ax.xaxis.set_minor_locator(
    ticker.LogLocator(base=10.0, subs=np.arange(2, 10), numticks=100)
)
ax.grid(True, which="minor", ls="-", lw=0.5, zorder=5)
ax.grid(True, which="major", ls="-", lw=1, zorder=10)
Dlo = production.sfd_range[0] * 1e-3
Dhi = production.sfd_range[1] * 1e-3
lo = Dvals < Dlo
hi = Dvals > Dhi
inrange = (Dvals >= Dlo) & (Dvals <= Dhi)
ax.plot(Dvals[inrange], Nvals[inrange], "-", color="black", linewidth=1.0, zorder=50)
ax.plot(Dvals[lo], Nvals[lo], "-.", color="orange", linewidth=2.0, zorder=50)
ax.plot(Dvals[hi], Nvals[hi], "-.", color="orange", linewidth=2.0, zorder=50)

plt.tight_layout()
plt.show()



import tempfile
import unittest

# This will suppress the warning issued by xarray starting in version 2023.12.0 about the change in the API regarding .dims
# The API change does not affect the functionality of the code, so we can safely ignore the warning
import warnings
from pathlib import Path

import numpy as np
import xarray as xr

import cratermaker
from cratermaker import Target
from cratermaker.constants import _COMBINED_DATA_FILE_NAME, _EXPORT_DIR

warnings.filterwarnings("ignore", category=FutureWarning, module="xarray")
warnings.filterwarnings("ignore", category=FutureWarning, module="uxarray")


def warning_with_breakpoint(message, category, filename, lineno, file=None, line=None):
    print(f"{category} Warning in line {lineno} of {filename} : {message}")
    return


warnings.simplefilter("always")  # Always trigger the warnings
warnings.showwarning = warning_with_breakpoint


class TestSimulation(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        target = Target(name="Moon")
        self.pix = target.radius / 10.0
        self.gridlevel = 5

    def test_simulation_defaults(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            self.assertEqual(sim.target.name, "Moon")

    def test_simulation_save(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # Test basic save operation
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            sim.save()

            # Test that variables are saved correctly
            sim.surface.update_elevation(1.0)
            np.testing.assert_array_equal(
                sim.surface.uxds["node_elevation"].values,
                np.ones(sim.surface.uxds.uxgrid.n_node),
            )
            np.testing.assert_array_equal(
                sim.surface.uxds["face_elevation"].values,
                np.ones(sim.surface.uxds.uxgrid.n_face),
            )

            sim.save()

            filename = Path(sim.data_dir) / _COMBINED_DATA_FILE_NAME.replace(".nc", f"{sim.interval_number:06d}.nc")
            self.assertTrue(filename.exists())
            with xr.open_dataset(filename) as ds:
                ds = ds.isel(time=-1)
                np.testing.assert_array_equal(ds["node_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_node))
                np.testing.assert_array_equal(ds["face_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_face))

            # Test saving combined data
            sim.save(combine_data_files=True)
            filename = _COMBINED_DATA_FILE_NAME

            filename = Path(sim.data_dir) / _COMBINED_DATA_FILE_NAME
            self.assertTrue(filename.exists())
            with xr.open_dataset(filename) as ds:
                ds = ds.isel(time=-1)
                np.testing.assert_array_equal(ds["node_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_node))
                np.testing.assert_array_equal(ds["face_elevation"].values, np.ones(sim.surface.uxds.uxgrid.n_face))

        return

    def test_emplace(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            cdiam = 2 * self.pix
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            crater = cratermaker.Crater.maker(final_diameter=cdiam)
            sim.emplace(crater)
            pdiam = crater.projectile_diameter

            sim.emplace(final_diameter=cdiam)
            sim.emplace(projectile_diameter=pdiam)
        return

    def test_populate(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            # Test that populate will work even if no craters are returned
            sim.populate(age=1e-6)
            sim.populate(age=3.8e3)
        return

    def test_run(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel, dosubpixel_degradation=True)
            sim.run(age=1000)

            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            sim.run(age=1000, age_interval=100)

            # Test that the simulation doesn't fail when the age doesn't divide evenly by the age_interval
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            sim.run(age=1010, age_interval=100)

    def test_invalid_run_args(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)

            # Test case: Neither the age nor the diameter_number argument is provided
            with self.assertRaises(ValueError):
                sim.run()

            # Test case: Both the age and diameter_number arguments are provided
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, diameter_number=(300e3, 80))

            # Test case: Both the age_end and diameter_number_end arguments are provided
            with self.assertRaises(ValueError):
                sim.run(age_end=3.0e3, diameter_number_end=(300e3, 80))

            # Test case: The age argument is provided but is not a scalar
            with self.assertRaises(ValueError):
                sim.run(age=[3.8e3])

            # Test case: The age_end argument is provided but is not a scalar
            with self.assertRaises(ValueError):
                sim.run(age_end=[3.0e3])

            # Test case: The age_interval is provided but is not a positive scalar
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, age_interval=-100.0)

            # Test case: The age_interval provided is negative, or is greater than age - age_end
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, age_end=3.0e3, age_interval=1000.0)

            # Test case: The diameter_number argument is not a pair of values, or any of them are less than 0
            with self.assertRaises(ValueError):
                sim.run(diameter_number=(300e3, -80))

            # Test case: The diameter_number_end argument is not a pair of values, or any of them are less than 0
            with self.assertRaises(ValueError):
                sim.run(diameter_number_end=(300e3, -80))

            # Test case: The diameter_number_interval argument is not a pair of values, or any of them are less than 0
            with self.assertRaises(ValueError):
                sim.run(diameter_number_interval=(300e3, -80))

            # Test case: The age_interval and diameter_number_interval arguments are both provided
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, age_interval=100.0, diameter_number_interval=(300e3, 80))

            # Test case: The diameter_number_interval provided is negative, or is greater than diameter_number - diameter_number_end
            with self.assertRaises(ValueError):
                sim.run(
                    diameter_number=(300e3, 80),
                    diameter_number_end=(300e3, 30),
                    diameter_number_interval=(300e3, 100),
                )

            # Test case: The ninterval is provided but is not an integer or is less than 1
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, ninterval=0)

            # Test case: The ninterval is provided and either age_interval or diameter_number_interval is also provided
            with self.assertRaises(ValueError):
                sim.run(age=3.8e3, ninterval=100, age_interval=100.0)

        return

    def test_simulation_to_config(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            # First simulation: no target passed, should default to "Moon"
            sim = cratermaker.Simulation(simdir=simdir, gridlevel=self.gridlevel)
            self.assertIsInstance(sim.target, Target)
            self.assertEqual(sim.target.name, "Moon")
            del sim

            # Second simulation: override target with "Mars"
            sim = cratermaker.Simulation(simdir=simdir, target="Mars", resume_old=True)
            self.assertEqual(sim.target.name, "Mars")
            del sim

            # Third simulation: no target passed, should read "Mars" from config
            sim = cratermaker.Simulation(simdir=simdir, resume_old=True)
            self.assertEqual(sim.target.name, "Mars")
            del sim
        return


if __name__ == "__main__":
    unittest.main()



import datetime
import inspect
import os
import sys

import cratermaker

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
project = "Cratermaker"
copyright = f"{datetime.datetime.now().year}, David A. Minton"
author = "David A. Minton"
version = cratermaker.__version__
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

sys.path.insert(0, os.path.abspath("_exts"))
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx_autosummary_accessors",
    "sphinx.ext.linkcode",
    "sphinxext.opengraph",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_inline_tabs",
    "sphinx_gallery.gen_gallery",
    "cratermaker_autodoc",
]

extlinks = {
    "issue": ("https://github.com/MintonGroup/cratermaker/issues/%s", "GH%s"),
    "pull": ("https://github.com/MintonGroup/cratermaker/pull/%s", "PR%s"),
    "discussion": ("https://github.com/MintonGroup/cratermaker/discussions/%s", "D%s"),
    "release": ("https://github.com/MintonGroup/cratermaker/releases/tag/%s", "%s"),
}


templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


# Use Autodoc and Napolean for extracting docstrings
autosummary_generate = True
autodoc_typehints = "none"
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_rtype = False

autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "show-inheritance": True,
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
}

templates_path = ["_templates"]

sphinx_gallery_conf = {
    "examples_dirs": "../examples",
    "gallery_dirs": "auto_examples",
}

html_theme = "sphinx_book_theme"
html_title = ""
html_static_path = ["_static"]

html_context = {
    "github_user": "MintonGroup",
    "github_repo": "cratermaker",
    "github_version": "main",
    "doc_path": "docs",
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = dict(
    # analytics_id=''  this is configured in rtfd.io
    # canonical_url="",
    repository_url="https://github.com/MintonGroup/cratermaker",
    repository_branch="main",
    navigation_with_keys=False,  # pydata/pydata-sphinx-theme#1492
    path_to_docs="docs",
    use_edit_page_button=True,
    use_repository_button=True,
    use_issues_button=True,
    home_page_in_toc=False,
    extra_footer="""<p>Development of Cratermaker was supported by NASA Lunar Data Analysis Program Grants <a href="https://www.usaspending.gov/award/ASST_NON_80NSSC21K1719_8000">#80NSSC21K1719</a> and <a href="https://www.usaspending.gov/award/ASST_NON_80NSSC25K7050_8000">#80NSSC25K7050</a><br>
    Theme by the <a href="https://ebp.jupyterbook.org">Executable Book Project</a></p>""",
    icon_links=[],  # workaround for pydata/pydata-sphinx-theme#1220
)


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logos/Cratermaker_Social_Preview.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/logos/Cratermaker_Icon.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["style.css"]

# configuration for sphinxext.opengraph
ogp_site_url = "https://cratermaker.readthedocs.io/en/latest/"
ogp_image = "https://cratermaker.readthedocs.io/en/stable/_static/logos/Cratermaker_Social_Preview.png"
ogp_custom_meta_tags = [
    '<meta name="image" property="og:image" content="https://cratermaker.readthedocs.io/en/stable/_static/logos/Cratermaker_Social_Preview.png" />',
]


# based on numpy doc/source/conf.py
def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        lineno = None

    if lineno:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname(cratermaker.__file__))


def html_page_context(app, pagename, templatename, context, doctree):
    # Disable edit button for docstring generated pages
    if "generated" in pagename:
        context["theme_use_edit_page_button"] = False



import unittest

from cratermaker import Target
from cratermaker.utils.general_utils import _create_catalogue


class TestTarget(unittest.TestCase):
    def test_target_from_catalogue(self):
        # Test creating a target from the built-in catalogue
        planets = ["Mercury", "Venus", "Earth", "Moon", "Mars", "Ceres", "Vesta"]
        for planet in planets:
            target = Target(name=planet)
            v = Target._catalogue[planet]
            self.assertEqual(target.radius, v["radius"])
            self.assertEqual(target.mass, v["mass"])
            self.assertEqual(target.material, v["material"])
            self.assertEqual(target.transition_scale_type, v["transition_scale_type"])
            self.assertIsInstance(target.radius, float)
            self.assertIsInstance(target.mass, float)
            self.assertIsInstance(target.material, str)
            self.assertIsInstance(target.transition_scale_type, str)

        # Test creating a target from a custom catalogue
        body_properties = [
            "name",
            "radius",
            "mass",
            "material",
            "transition_scale_type",
        ]
        body_values = [
            ("Arrakis", 5995.0e3, 4.98e24, "Sand", "silicate"),
            ("Salusa Secundus", 7200.0e3, 8.62e24, "Hard Rock", "silicate"),
        ]
        catalogue = _create_catalogue(body_properties, body_values)
        for planet, v in catalogue.items():
            target = Target(name=planet, catalogue=catalogue)
            self.assertEqual(target.radius, v["radius"])
            self.assertEqual(target.mass, v["mass"])
            self.assertEqual(target.material, v["material"])
            self.assertEqual(target.transition_scale_type, v["transition_scale_type"])
            self.assertIsInstance(target.radius, float)
            self.assertIsInstance(target.mass, float)

    def test_custom_target(self):
        # Test creating a custom target
        target = Target(
            name="Arrakis",
            radius=5995.0e3,
            mass=4.98e24,
            material="Sand",
            transition_scale_type="silicate",
        )
        self.assertEqual(target.radius, 5995.0e3)
        self.assertEqual(target.mass, 4.98e24)
        self.assertEqual(target.material, "Sand")

    def test_invalid_target(self):
        # Test incomplete custom target creation or invalid arguments
        with self.assertRaises(ValueError):
            Target("Arrakis")

        # Missing name
        with self.assertRaises(TypeError):
            Target(
                radius=5995.0,
                mass=4.98e24,
                material="Sand",
                transition_scale_type="silicate",
            )
        with self.assertRaises(ValueError):
            Target("Moon", transition_scale_type="flubber")


if __name__ == "__main__":
    unittest.main()



def process_component_docstring(app, what, name, obj, options, lines):
    from cratermaker.components.scaling import Scaling
    from cratermaker.components.production import Production
    from cratermaker.components.morphology import Morphology
    from cratermaker.components.surface import Surface
    from cratermaker.components.projectile import Projectile

    component_classes = {
        "Scaling": Scaling,
        "Production": Production,
        "Morphology": Morphology,
        "Surface": Surface,
        "Projectile": Projectile,
    }

    if name.endswith('.maker'):
        parent_name = name.rsplit('.', 1)[0]
        cls_name = parent_name.split('.')[-1].capitalize()
        cls = component_classes.get(cls_name)
        if cls:
            try:
                available = cls.available()
                if available:
                    lines.append("")
                    lines.append("**Available models:**")
                    lines.append("")
                    for item in available:
                        lines.append(f"- ``{item}``")
            except Exception as e:
                lines.append("")
                lines.append(f"*Error listing available components: {e}*")

def setup(app):
    app.connect('autodoc-process-docstring', process_component_docstring)


from __future__ import annotations

import importlib
import pkgutil
from abc import ABC
from typing import Any

from cratermaker.core.base import CratermakerBase
from cratermaker.utils.general_utils import parameter


class ComponentBase(CratermakerBase, ABC):
    _registry: dict[str, type[ComponentBase]] = {}

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def __str__(self) -> str:
        # Get the name of the class
        base_class = type(self).__mro__[1].__name__
        return f"<{base_class}: {self._component_name}>"

    @classmethod
    def maker(
        cls,
        component: str | type[ComponentBase] | ComponentBase | None = None,
        **kwargs: Any,
    ) -> ComponentBase:
        """
        Initialize a component model with the given name or instance.

        Parameters
        ----------
        component : str or ComponentBase or None
            The name of the component to use, or an instance of ComponentBase. If None, it choose a default component.
        kwargs : Any
            Additional keyword arguments to pass to the component model constructor.

        Returns
        -------
        component
            An instance of the specified component model.

        Raises
        ------
        KeyError
            If the specified component model name is not found in the registry.
        TypeError
            If the specified component model is not a string or a subclass of component.
        """

        if component is None:
            component = cls.available()[0]  # Default to the first available component
        if isinstance(component, str):
            if component not in cls.available():
                raise KeyError(
                    f"Unknown component model: {component}. Available models: {cls.available()}"
                )
            return cls._registry[component](**kwargs)
        elif isinstance(component, type) and issubclass(component, ComponentBase):
            return component(**kwargs)
        elif isinstance(component, ComponentBase):
            return component
        else:
            raise TypeError(
                f"component must be a string or a subclass of component, not {type(component)}"
            )

    @parameter
    def name(self):
        """
        The registered name of this scaling model set by the @ComponentBase.register decorator.
        """
        return self._component_name

    @classmethod
    def register(cls, name: str):
        """
        Class decorator to register a component model component under the given key.
        """

        def decorator(subcls):
            subcls._component_name = name
            subcls._registry[name] = subcls
            return subcls

        return decorator

    @classmethod
    def available(cls) -> list[str]:
        """Return list of all registered catalogue names."""
        return list(cls._registry.keys())


def import_components(
    package_name: str, package_path: list[str], ignore_private: bool = True
) -> None:
    """
    Import all modules of a component package, optionally skipping private modules.

    Parameters
    ----------
    package_name : str
        The full name of the package (e.g., "cratermaker.components.component").
    package_path : list[str]
        The __path__ attribute of the package (usually just __path__).
    ignore_private : bool, optional
        Whether to ignore modules whose names start with an underscore (default is True).
    """
    for _, module_name, _ in pkgutil.iter_modules(package_path):
        if ignore_private and module_name.startswith("_"):
            continue
        importlib.import_module(f"{package_name}.{module_name}")



"""
Plot the NPF Chronology function for the Moon and Mars
=======================================================

.. rubric:: By Austin Blevins and David Minton

Here we plot the chronology component of the Neukum production function for the Moon and Mars. The chronology function is the number of craters per unit area per unit time, which is a function of crater diameter and age. The plot shows the number of craters per square kilometer per million years as a function of time for a 1 km crater.
"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker import Production

moon = Production.maker("neukum", version="Moon")
mars = Production.maker("neukum", version="Mars")
age_vals = np.linspace(4.5e3, 0.0, num=1000)

N1_moon = moon.chronology(age=age_vals)
N1_mars = mars.chronology(age=age_vals)

fig, ax = plt.subplots(figsize=(8, 4))
ax.set_yscale("log")
ax.set_ylabel("$\\mathregular{N(1) (km^{-2})}$")
ax.set_xlabel("Time (Gy ago)")
ax.set_xlim(4.5, 0)

# Convert age from My to Gy and N1 from m^2 to km^2 for plotting
ax.plot(
    age_vals * 1e-3,
    N1_moon * 1e6,
    "-",
    color="dimgrey",
    linewidth=2.0,
    zorder=50,
    label="Moon",
)
ax.plot(
    age_vals * 1e-3,
    N1_mars * 1e6,
    "-",
    color="orange",
    linewidth=2.0,
    zorder=50,
    label="Mars",
)

ax.legend()

plt.tight_layout()
plt.show()



import tempfile
import unittest

import numpy as np

from cratermaker import Crater, Simulation
from cratermaker.components.morphology import Morphology
from cratermaker.components.surface import Surface
from cratermaker.components.target import Target

morphology_models = Morphology.available()


class TestMorphology(unittest.TestCase):
    def setUp(self):
        # Initialize a target and surface for testing
        self.target = Target(name="Moon")
        self.pix = self.target.radius / 10.0
        self.gridlevel = 5

        self.dummy_crater = Crater.maker(
            location=(0.0, 0.0),
            final_diameter=100000.0,
        )

    def test_model_registration(self):
        models = Morphology.available()
        self.assertIn("simplemoon", models)

    def test_model_instantiation(self):
        for model_name in morphology_models:
            morphology = Morphology.maker(model_name)
            self.assertIsInstance(morphology, object)
            self.assertEqual(morphology.name, model_name)
            morphology.crater = self.dummy_crater
            self.assertIs(morphology.crater, self.dummy_crater)

    def test_form_crater_executes(self):
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            surface = Surface.maker(simdir=simdir, target=self.target, reset=True, gridlevel=self.gridlevel)
            for model_name in morphology_models:
                morphology = Morphology.maker(model_name, surface=surface)
                morphology.emplace(self.dummy_crater)

    def test_make_morphology(self):
        # Test the make_morphology function
        for model_name in morphology_models:
            morphology = Morphology.maker(moprhology=model_name)
            self.assertIsInstance(morphology, Morphology)

    def test_finite_profile_values(self):
        for model_name in morphology_models:
            morphology = Morphology.maker(model_name)
            crater_radius_values = [1.0, 1e3, 15e3, 50e3, 500e3, 3000e3]
            rvals = np.linspace(0, 10, 1000)
            for final_radius in crater_radius_values:
                crater = Crater.maker(final_radius=final_radius)
                crater_shape = morphology.crater_profile(crater, rvals * final_radius)
                self.assertTrue(
                    np.all(np.isfinite(crater_shape)),
                    f"Crater profile for {model_name} contains NaN or Inf values.",
                )
                ejecta_shape = morphology.ejecta_profile(crater, rvals * final_radius)
                self.assertTrue(
                    np.all(np.isfinite(ejecta_shape)),
                    f"Ejecta profile for {model_name} contains NaN or Inf values.",
                )

    def test_crater_depth_surface(self):
        from cratermaker.components.morphology.simplemoon import SimpleMoonCrater
        # Tests that the surface elevations are expected

        final_diameter_list = [100e3, 200e3, 500e3, 1000e3]
        delta_vals = [0.4, 0.3, 0.3, 0.2]

        gridargs = {
            "icosphere": {"gridlevel": 6},
            "arbitrary_resolution": {
                "pix": 30e3,
            },
            "hireslocal": {
                "pix": 1.0e3,
                "local_location": (0, 0),
                "local_radius": 100e3,
            },
        }

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as simdir:
            for name, args in gridargs.items():
                sim = Simulation(simdir=simdir, surface=name, **args)
                for final_diameter, delta in zip(final_diameter_list, delta_vals, strict=False):
                    sim.surface.reset()
                    # verify that the surface is flat
                    self.assertAlmostEqual(
                        sim.surface.node_elevation.min(),
                        0.0,
                        delta=1e0,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )
                    self.assertAlmostEqual(
                        sim.surface.face_elevation.min(),
                        0.0,
                        delta=1e0,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )
                    self.assertAlmostEqual(
                        sim.surface.node_elevation.max(),
                        0.0,
                        delta=1e0,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )
                    self.assertAlmostEqual(
                        sim.surface.face_elevation.max(),
                        0.0,
                        delta=1e0,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )

                    crater = SimpleMoonCrater.maker(final_diameter=final_diameter, location=(0, 0))
                    sim.emplace(crater)

                    # Verify that the crater depth and rim heights are close to the expected values
                    self.assertAlmostEqual(
                        -sim.surface.node_elevation.min() / crater.floor_depth,
                        1.0,
                        delta=delta,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )
                    self.assertAlmostEqual(
                        -sim.surface.face_elevation.min() / crater.floor_depth,
                        1.0,
                        delta=delta,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )
                    self.assertAlmostEqual(
                        sim.surface.node_elevation.max() / crater.rim_height,
                        1.0,
                        delta=2 * delta,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )
                    self.assertAlmostEqual(
                        sim.surface.face_elevation.max() / crater.rim_height,
                        1.0,
                        delta=2 * delta,
                        msg=f"Failed for {name} with diameter {final_diameter}",
                    )


if __name__ == "__main__":
    unittest.main()



"""
Plot random projectile impact angles
====================================

.. rubric:: By David Minton and Dennise Valadez

This example demonstrates how to generate and visualize random impact angles for projectiles using cratermaker.

"""

import matplotlib.pyplot as plt
import numpy as np

from cratermaker import Projectile

# Sample data generation
proj = Projectile(mean_velocity=5000, density=3000)
size = 10000
angles = np.array([proj.new_projectile().angle for _ in range(size)])

# Number of bins
bins = 50
observed_counts, bins_ang = np.histogram(angles, bins=bins, range=(0.0, 90.0))

# Expected distribution
uniform_dist = np.linspace(0, 1, size)
transformed_angles = np.rad2deg(np.arcsin(np.sqrt(uniform_dist)))
expected_counts, _ = np.histogram(transformed_angles, bins=bins, range=(0.0, 90.0))

# Plotting distribution
fig, ax = plt.subplots(figsize=(8, 4))
ax.bar(
    bins_ang[:-1],
    observed_counts,
    width=np.diff(bins_ang),
    align="edge",
    label="Observed",
    alpha=0.5,
)
ax.plot(bins_ang[:-1], expected_counts, label="Expected", color="red")
ax.set_xlabel("Impact Angle (deg)")
ax.set_ylabel("Count")
ax.legend()
ax.set_title("Impact Angle Distribution")
plt.tight_layout()
plt.show()



import numpy as np

# Custom type aliases for better readability
FloatLike = float | int | np.number
PairOfFloats = tuple[float, float] | list[float] | np.ndarray

# Default filenames and paths
_CONFIG_FILE_NAME = "cratermaker.yaml"
_SURFACE_DIR = "surface_data"
_CRATER_DIR = "crater_data"
_EXPORT_DIR = "export"
_COMBINED_DATA_FILE_NAME = "surface.nc"
_GRID_FILE_NAME = "grid.nc"
_CIRCLE_FILE_NAME = "circles.vtp"
_VTK_FILE_EXTENSION = "vtp"
_COMPONENT_NAMES = [
    "target",
    "scaling",
    "production",
    "morphology",
    "projectile",
    "surface",
]

# This is a factor used to determine the smallest length scale in the grid
_SMALLFAC = 1.0e-5
_VSMALL = 10 * np.finfo(np.float64).tiny
_LOGVSMALL = np.log10(_VSMALL)


# Optional: controlled public API
__all__ = []



from collections.abc import Sequence
from typing import Any

import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike, NDArray

from cratermaker.components.production import Production
from cratermaker.constants import FloatLike


@Production.register("powerlaw")
class PowerLawProduction(Production):
    """
    An operations class for computing the production function for craters and projectiles. This impliments a very simple power law
    production function that can be used as either a crater or projectile production function. The production function is defined as
    the cumulative number of craters greater than a given diameter per unit m^2 surface area.

    Parameters
    ----------
    generator_type : str, optional
        The type of generator to use. This can be either "crater" or "projectile". Default is "crater".
    N1_coef : float, optional
        The coefficient for the power law production function at 1 m diameter per 1 My.
        Defaults to 7.9.e-3 (lunar craters) or 2.2e-8 (lunar projectiles) based on fits to the NPF on the Moon.
    slope : float, optional
        The slope of the power law production function.
        Defaults to -3.33 (lunar craters) or -2.26 (lunar projectiles) based on fits to the NPF on the Moon.
    rng : numpy.random.Generator | None
        A numpy random number generator. If None, a new generator is created using the rng_seed if it is provided.
    rng_seed : Any type allowed by the rng_seed argument of numpy.random.Generator, optional
        The rng_rng_seed for the RNG. If None, a new RNG is created.
    rng_state : dict, optional
        The state of the random number generator. If None, a new state is created.
    **kwargs : Any
        Additional keyword arguments.
    """

    def __init__(
        self,
        generator_type: str = "crater",
        N1_coef: FloatLike | None = None,
        slope: FloatLike | None = None,
        rng: Generator | None = None,
        rng_seed: int | None = None,
        rng_state: dict | None = None,
        **kwargs: Any,
    ):
        super().__init__(rng=rng, rng_seed=rng_seed, rng_state=rng_state, **kwargs)
        self.generator_type = generator_type

        # Default values that are approximately equal to the NPF for the Moon
        default_N1_coef = {"crater": 7.883e-3, "projectile": 7.989e-7}

        default_slope = {"crater": -3.328, "projectile": -2.634}

        # Set the power law parameters for the production function along with defaults
        if N1_coef is None:
            N1_coef = default_N1_coef[self.generator_type]

        if not isinstance(N1_coef, FloatLike):
            raise ValueError("N1_coef must be a float")
        if N1_coef < 0.0:
            raise ValueError("N1_coef must be positive")
        self.N1_coef = N1_coef

        # Set the power law exponent for the production function along with defaults
        if slope is None:
            slope = default_slope[self.generator_type]

        if not isinstance(slope, FloatLike):
            raise ValueError("slope must be a float")
        elif (
            slope > 0.0
        ):  # Slope must be negative, but convention in the field is mixed. So we flip the sign if it is positive.
            slope *= -1
        self.slope = slope

    def __str__(self) -> str:
        base = super().__str__()
        return f"{base}\nN1 Coefficient: {self.N1_coef:.2e}\nSlope: {self.slope:.3f}"

    def function(
        self,
        diameter: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        age: FloatLike | Sequence[FloatLike] | ArrayLike = 1.0,
        age_end: FloatLike | Sequence[FloatLike] | ArrayLike | None = None,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> FloatLike | ArrayLike:
        """
        Return the cumulative size-frequency distribution of craters over a given age range and crater diameter for a simple power
        law model.

        Parameters
        ----------
        diameter : FloatLike or ArrayLike
            Crater diameter(s) in units of meters to compute corresponding cumulative number density value.
        age : FloatLike or ArrayLike, default=1.0
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        age_end, FloatLike or ArrayLike, optional
            The ending age in units of My relative to the present, which is used to compute the cumulative SFD. The default is 0 (present day).
        validate_inputs: bool, default=True
            If True, the function will check that the validity of age, age_end, and diameter arguments. If False, no check is performed, and the arguments are assumed to be valid. This can be used to speed up the function, particularly if it is called as part of a solver or optimization routine where the inputs are already known to be valid.
        **kwargs : Any
            Any additional keywords. These are not used in this base class, but included here so that any extended class can share
            the same function signature.

        Returns
        -------
        FloatLike or numpy array of FloatLike
            The cumulative number of craters per square meter greater than the input diameter that would be expected to form on a
            surface over the given age range.
        """
        if validate_inputs:
            diameter, _ = self._validate_csfd(diameter=diameter)
            age, age_end = self._validate_age(age, age_end)
        elif age_end is None:
            if np.isscalar(age):
                age_end = 0.0
            else:
                age_end = np.zeros_like(age)

        n_array = np.asarray(self.csfd(diameter))
        age_difference = np.asarray(age - age_end)

        if n_array.ndim > 0 and age_difference.ndim > 0:
            return n_array[:, None] * age_difference
        else:
            return n_array * age_difference

    def chronology(
        self,
        age: ArrayLike = np.array([1000.0]),
        age_end: ArrayLike | None = None,
        validate_inputs: bool = True,
        **kwargs: Any,
    ) -> NDArray[np.float64]:
        """
        Returns the age in My. Because the powerlaw model assumes constant impact rate, the returned age is the same as the input age.

        Parameters
        ----------
        age : FloatLike or ArrayLike, default=1.0
            Age in the past relative to the present day to compute cumulative SFD in units of My.
        **kwargs: Any
            Any additional keywords that are passed to the function method.

        Returns
        -------
        NDArray
            The number of craters relative to the amount produced in the last 1 My.

        """
        if validate_inputs:
            age, age_end = self._validate_age(age, age_end)
        else:
            if age_end is None:
                if np.isscalar(age):
                    age_end = 0.0
                else:
                    age_end = np.zeros_like(age)

        return age - age_end

    def csfd(
        self, diameter: FloatLike | ArrayLike, **kwargs: Any
    ) -> FloatLike | ArrayLike:
        """
        Return the cumulative size frequency distribution of craters at a given age relative to age = 1 My ago per m^2.

        Parameters
        ----------
        diameter : FloatLike or ArrayLike
            units of meter

        Returns
        -------
        FloatLike or numpy array
           Cumulative number density of per square meter greater than the input diameter.
        """
        return self.N1_coef * diameter**self.slope

    @property
    def N1_coef(self):
        """Get the N1 coefficient of the power law production function."""
        return self._N1_coef

    @N1_coef.setter
    def N1_coef(self, value):
        """Set the N1 coefficient of the power law production function."""
        if not isinstance(value, FloatLike):
            raise TypeError("N1_coef must be a numeric value (float or int)")
        if value < 0:
            raise ValueError("N1_coef must be positive")
        self._N1_coef = value

    @property
    def slope(self):
        """Get the slope of the power law production function."""
        return self._slope

    @slope.setter
    def slope(self, value):
        """Set the slope of the power law production function."""
        if not isinstance(value, FloatLike):
            raise TypeError("slope must be a numeric value (float or int)")
        self._slope = value
